#!/usr/bin/env python3
import json
import os.path as path
from os import makedirs, listdir, remove
from deemix import __version__ as deemixVersion
import logging
import datetime
import platform

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('deemix')

import deemix.utils.localpaths as localpaths

class Settings:
    def __init__(self, configFolder=None):
        self.settings = {}
        self.configFolder = configFolder
        if not self.configFolder:
            self.configFolder = localpaths.getConfigFolder()
        self.defaultSettings = {
          "downloadLocation": path.join(localpaths.getHomeFolder(), 'deemix Music'),
          "tracknameTemplate": "%artist% - %title%",
          "albumTracknameTemplate": "%tracknumber% - %title%",
          "playlistTracknameTemplate": "%position% - %artist% - %title%",
          "createPlaylistFolder": True,
          "playlistNameTemplate": "%playlist%",
          "createArtistFolder": False,
          "artistNameTemplate": "%artist%",
          "createAlbumFolder": True,
          "albumNameTemplate": "%artist% - %album%",
          "createCDFolder": True,
          "createStructurePlaylist": False,
          "createSingleFolder": False,
          "padTracks": True,
          "paddingSize": "0",
          "illegalCharacterReplacer": "_",
          "queueConcurrency": 3,
          "maxBitrate": "3",
          "fallbackBitrate": True,
          "fallbackSearch": False,
          "logErrors": True,
          "logSearched": False,
          "saveDownloadQueue": False,
          "overwriteFile": "n",
          "createM3U8File": False,
          "playlistFilenameTemplate": "playlist",
          "syncedLyrics": False,
          "embeddedArtworkSize": 800,
          "embeddedArtworkPNG": False,
          "localArtworkSize": 1400,
          "localArtworkFormat": "jpg",
          "saveArtwork": True,
          "coverImageTemplate": "cover",
          "saveArtworkArtist": False,
          "artistImageTemplate": "folder",
          "jpegImageQuality": 80,
          "dateFormat": "Y-M-D",
          "albumVariousArtists": True,
          "removeAlbumVersion": False,
          "removeDuplicateArtists": False,
          "featuredToTitle": "0",
          "titleCasing": "nothing",
          "artistCasing": "nothing",
          "executeCommand": "",
          "tags": {
            "title": True,
            "artist": True,
            "album": True,
            "cover": True,
            "trackNumber": True,
            "trackTotal": False,
            "discNumber": True,
            "discTotal": False,
            "albumArtist": True,
            "genre": True,
            "year": True,
            "date": True,
            "explicit": False,
            "isrc": True,
            "length": True,
            "barcode": True,
            "bpm": True,
            "replayGain": False,
            "label": True,
            "lyrics": False,
            "copyright": False,
            "composer": False,
            "involvedPeople": False,
            "savePlaylistAsCompilation": False,
            "useNullSeparator": False,
            "saveID3v1": True,
            "multiArtistSeparator": "default",
            "singleAlbumArtist": False
          }
        }

        # Create config folder if it doesn't exsist
        makedirs(self.configFolder, exist_ok=True)

        # Create config file if it doesn't exsist
        if not path.isfile(path.join(self.configFolder, 'config.json')):
            with open(path.join(self.configFolder, 'config.json'), 'w') as f:
                json.dump(self.defaultSettings, f, indent=2)

        # Read config file
        with open(path.join(self.configFolder, 'config.json'), 'r') as configFile:
            self.settings = json.load(configFile)

        self.settingsCheck()

        # Make sure the download path exsits
        makedirs(self.settings['downloadLocation'], exist_ok=True)

        # LOGFILES

        # Create logfile name and path
        logspath = path.join(self.configFolder, 'logs')
        now = datetime.datetime.now()
        logfile = now.strftime("%Y-%m-%d_%H%M%S")+".log"
        makedirs(logspath, exist_ok=True)

        # Add handler for logging
        fh = logging.FileHandler(path.join(logspath, logfile), 'w', 'utf-8')
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(logging.Formatter('%(asctime)s - [%(levelname)s] %(message)s'))
        logger.addHandler(fh)
        logger.info(f"{platform.platform(True, True)} - Python {platform.python_version()}, deemix {deemixVersion}")

        # Only keep last 5 logfiles (to preserve disk space)
        logslist = listdir(logspath)
        logslist.sort()
        if len(logslist)>5:
            for i in range(len(logslist)-5):
                remove(path.join(logspath, logslist[i]))

    # Saves the settings
    def saveSettings(self, newSettings=None):
        if newSettings:
            self.settings = newSettings
        with open(path.join(self.configFolder, 'config.json'), 'w') as configFile:
            json.dump(self.settings, configFile, indent=2)

    # Checks if the default settings have changed
    def settingsCheck(self):
        changes = 0
        for x in self.defaultSettings:
            if not x in self.settings or type(self.settings[x]) != type(self.defaultSettings[x]):
                self.settings[x] = self.defaultSettings[x]
                changes += 1
        for x in self.defaultSettings['tags']:
            if not x in self.settings['tags'] or type(self.settings['tags'][x]) != type(self.defaultSettings['tags'][x]):
                self.settings['tags'][x] = self.defaultSettings['tags'][x]
                changes += 1
        if self.settings['downloadLocation'] == "":
            self.settings['downloadLocation'] = path.join(localpaths.getHomeFolder(), 'deemix Music')
            changes += 1
        for template in ['tracknameTemplate', 'albumTracknameTemplate', 'playlistTracknameTemplate', 'playlistNameTemplate', 'artistNameTemplate', 'albumNameTemplate', 'playlistFilenameTemplate', 'coverImageTemplate', 'artistImageTemplate']:
            if self.settings[template] == "":
                self.settings[template] = self.defaultSettings[template]
                changes += 1
        if changes > 0:
            self.saveSettings()



#!/usr/bin/env python3
import os.path
import re

from requests import get
from requests import exceptions as request_exception

from concurrent.futures import ThreadPoolExecutor
from os import makedirs, remove, system as execute
from tempfile import gettempdir
from time import sleep

from deemix.app.queueitem import QIConvertable, QISingle, QICollection
from deemix.app.track import Track
from deemix.utils.misc import changeCase
from deemix.utils.pathtemplates import generateFilename, generateFilepath, settingsRegexAlbum, settingsRegexArtist, settingsRegexPlaylistFile
from deemix.api.deezer import USER_AGENT_HEADER
from deemix.utils.taggers import tagID3, tagFLAC

from Cryptodome.Cipher import Blowfish
from mutagen.flac import FLACNoHeaderError
import logging

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('deemix')

TEMPDIR = os.path.join(gettempdir(), 'deemix-imgs')
if not os.path.isdir(TEMPDIR):
    makedirs(TEMPDIR)

extensions = {
    9: '.flac',
    0: '.mp3',
    3: '.mp3',
    1: '.mp3',
    8: '.mp3',
    15: '.mp4',
    14: '.mp4',
    13: '.mp4'
}

errorMessages = {
    'notOnDeezer': "Track not available on Deezer!",
    'notEncoded': "Track not yet encoded!",
    'notEncodedNoAlternative': "Track not yet encoded and no alternative found!",
    'wrongBitrate': "Track not found at desired bitrate.",
    'wrongBitrateNoAlternative': "Track not found at desired bitrate and no alternative found!",
    'no360RA': "Track is not available in Reality Audio 360.",
    'notAvailable': "Track not available on deezer's servers!",
    'notAvailableNoAlternative': "Track not available on deezer's servers and no alternative found!"
}
def downloadImage(url, path, overwrite="n"):
    if not os.path.isfile(path) or overwrite in ['y', 't', 'b']:
        try:
            image = get(url, headers={'User-Agent': USER_AGENT_HEADER}, timeout=30)
            image.raise_for_status()
            with open(path, 'wb') as f:
                f.write(image.content)
            return path
        except request_exception.HTTPError:
            if 'cdns-images.dzcdn.net' in url:
                urlBase = url[:url.rfind("/")+1]
                pictureUrl = url[len(urlBase):]
                pictureSize = int(pictureUrl[:pictureUrl.find("x")])
                if pictureSize > 1200:
                    logger.warn("Couldn't download "+str(pictureSize)+"x"+str(pictureSize)+" image, falling back to 1200x1200")
                    sleep(1)
                    return  downloadImage(urlBase+pictureUrl.replace(str(pictureSize)+"x"+str(pictureSize), '1200x1200'), path, overwrite)
            logger.error("Couldn't download Image: "+url)
        except:
            sleep(1)
            return downloadImage(url, path, overwrite)
        if os.path.isfile(path): remove(path)
        return None
    else:
        return path

def formatDate(date, template):
    elements = {
        'year': ['YYYY', 'YY', 'Y'],
        'month': ['MM', 'M'],
        'day': ['DD', 'D']
    }
    for element, placeholders in elements.items():
        for placeholder in placeholders:
            if placeholder in template:
                template = template.replace(placeholder, str(date[element]))
    return template

class DownloadJob:
    def __init__(self, dz, sp, queueItem, interface=None):
        self.dz = dz
        self.sp = sp
        self.interface = interface
        if isinstance(queueItem, QIConvertable) and queueItem.extra:
            self.sp.convert_spotify_playlist(self.dz, queueItem, interface=self.interface)
        self.queueItem = queueItem
        self.settings = queueItem.settings
        self.bitrate = queueItem.bitrate
        self.downloadPercentage = 0
        self.lastPercentage = 0
        self.extrasPath = None
        self.playlistPath = None
        self.playlistURLs = []

    def start(self):
        if not self.queueItem.cancel:
            if isinstance(self.queueItem, QISingle):
                result = self.downloadWrapper(self.queueItem.single)
                if result:
                    self.singleAfterDownload(result)
            elif isinstance(self.queueItem, QICollection):
                tracks = [None] * len(self.queueItem.collection)
                with ThreadPoolExecutor(self.settings['queueConcurrency']) as executor:
                    for pos, track in enumerate(self.queueItem.collection, start=0):
                        tracks[pos] = executor.submit(self.downloadWrapper, track)
                self.collectionAfterDownload(tracks)
        if self.interface:
            if self.queueItem.cancel:
                self.interface.send('currentItemCancelled', self.queueItem.uuid)
                self.interface.send("removedFromQueue", self.queueItem.uuid)
            else:
                self.interface.send("finishDownload", self.queueItem.uuid)
        return self.extrasPath

    def singleAfterDownload(self, result):
        if not self.extrasPath:
            self.extrasPath = self.settings['downloadLocation']
        # Save Album Cover
        if self.settings['saveArtwork'] and 'albumPath' in result:
            for image in result['albumURLs']:
                downloadImage(image['url'], f"{result['albumPath']}.{image['ext']}", self.settings['overwriteFile'])
        # Save Artist Artwork
        if self.settings['saveArtworkArtist'] and 'artistPath' in result:
            for image in result['artistURLs']:
                downloadImage(image['url'], f"{result['artistPath']}.{image['ext']}", self.settings['overwriteFile'])
        # Create searched logfile
        if self.settings['logSearched'] and 'searched' in result:
            with open(os.path.join(self.extrasPath, 'searched.txt'), 'wb+') as f:
                orig = f.read().decode('utf-8')
                if not result['searched'] in orig:
                    if orig != "":
                        orig += "\r\n"
                    orig += result['searched'] + "\r\n"
                f.write(orig.encode('utf-8'))
        # Execute command after download
        if self.settings['executeCommand'] != "":
            execute(self.settings['executeCommand'].replace("%folder%", self.extrasPath).replace("%filename%", result['filename']))

    def collectionAfterDownload(self, tracks):
        if not self.extrasPath:
            self.extrasPath = self.settings['downloadLocation']
        playlist = [None] * len(tracks)
        errors = ""
        searched = ""

        for index in range(len(tracks)):
            result = tracks[index].result()
            # Check if queue is cancelled
            if not result:
                return None
            # Log errors to file
            if 'error' in result:
                if not 'data' in result['error']:
                    result['error']['data'] = {'id': "0", 'title': 'Unknown', 'artist': 'Unknown'}
                errors += f"{result['error']['data']['id']} | {result['error']['data']['artist']} - {result['error']['data']['title']} | {result['error']['message']}\r\n"
            # Log searched to file
            if 'searched' in result:
                searched += result['searched'] + "\r\n"
            # Save Album Cover
            if self.settings['saveArtwork'] and 'albumPath' in result:
                for image in result['albumURLs']:
                    downloadImage(image['url'], f"{result['albumPath']}.{image['ext']}", self.settings['overwriteFile'])
            # Save Artist Artwork
            if self.settings['saveArtworkArtist'] and 'artistPath' in result:
                for image in result['artistURLs']:
                    downloadImage(image['url'], f"{result['artistPath']}.{image['ext']}", self.settings['overwriteFile'])
            # Save filename for playlist file
            playlist[index] = ""
            if 'filename' in result:
                playlist[index] = result['filename']

        # Create errors logfile
        if self.settings['logErrors'] and errors != "":
            with open(os.path.join(self.extrasPath, 'errors.txt'), 'wb') as f:
                f.write(errors.encode('utf-8'))
        # Create searched logfile
        if self.settings['logSearched'] and searched != "":
            with open(os.path.join(self.extrasPath, 'searched.txt'), 'wb') as f:
                f.write(searched.encode('utf-8'))
        # Save Playlist Artwork
        if self.settings['saveArtwork'] and self.playlistPath and not self.settings['tags']['savePlaylistAsCompilation']:
            for image in self.playlistURLs:
                downloadImage(image['url'], os.path.join(self.extrasPath, self.playlistPath)+f".{image['ext']}", self.settings['overwriteFile'])
        # Create M3U8 File
        if self.settings['createM3U8File']:
            filename = settingsRegexPlaylistFile(self.settings['playlistFilenameTemplate'], self.queueItem, self.settings) or "playlist"
            with open(os.path.join(self.extrasPath, filename+'.m3u8'), 'wb') as f:
                for line in playlist:
                    f.write((line + "\n").encode('utf-8'))
        # Execute command after download
        if self.settings['executeCommand'] != "":
            execute(self.settings['executeCommand'].replace("%folder%", self.extrasPath))

    def download(self, trackAPI_gw, track=None):
        result = {}
        if self.queueItem.cancel: raise DownloadCancelled

        if trackAPI_gw['SNG_ID'] == "0":
            raise DownloadFailed("notOnDeezer")

        # Create Track object
        if not track:
            logger.info(f"[{trackAPI_gw['ART_NAME']} - {trackAPI_gw['SNG_TITLE']}] Getting the tags")
            track = Track(self.dz,
                          settings=self.settings,
                          trackAPI_gw=trackAPI_gw,
                          trackAPI=trackAPI_gw['_EXTRA_TRACK'] if '_EXTRA_TRACK' in trackAPI_gw else None,
                          albumAPI=trackAPI_gw['_EXTRA_ALBUM'] if '_EXTRA_ALBUM' in trackAPI_gw else None
                          )
            if self.queueItem.cancel: raise DownloadCancelled

        if track.MD5 == '':
            if track.fallbackId != "0":
                logger.warn(f"[{track.mainArtist['name']} - {track.title}] Track not yet encoded, using fallback id")
                newTrack = self.dz.get_track_gw(track.fallbackId)
                track.parseEssentialData(self.dz, newTrack)
                return self.download(trackAPI_gw, track)
            elif not track.searched and self.settings['fallbackSearch']:
                logger.warn(f"[{track.mainArtist['name']} - {track.title}] Track not yet encoded, searching for alternative")
                searchedId = self.dz.get_track_from_metadata(track.mainArtist['name'], track.title, track.album['title'])
                if searchedId != "0":
                    newTrack = self.dz.get_track_gw(searchedId)
                    track.parseEssentialData(self.dz, newTrack)
                    track.searched = True
                    return self.download(trackAPI_gw, track)
                else:
                    raise DownloadFailed("notEncodedNoAlternative")
            else:
                raise DownloadFailed("notEncoded")

        selectedFormat = self.getPreferredBitrate(track)
        if selectedFormat == -100:
            if track.fallbackId != "0":
                logger.warn(f"[{track.mainArtist['name']} - {track.title}] Track not found at desired bitrate, using fallback id")
                newTrack = self.dz.get_track_gw(track.fallbackId)
                track.parseEssentialData(self.dz, newTrack)
                return self.download(trackAPI_gw, track)
            elif not track.searched and self.settings['fallbackSearch']:
                logger.warn(f"[{track.mainArtist['name']} - {track.title}] Track not found at desired bitrate, searching for alternative")
                searchedId = self.dz.get_track_from_metadata(track.mainArtist['name'], track.title, track.album['title'])
                if searchedId != "0":
                    newTrack = self.dz.get_track_gw(searchedId)
                    track.parseEssentialData(self.dz, newTrack)
                    track.searched = True
                    return self.download(trackAPI_gw, track)
                else:
                    raise DownloadFailed("wrongBitrateNoAlternative")
            else:
                raise DownloadFailed("wrongBitrate")
        elif selectedFormat == -200:
            raise DownloadFailed("no360RA")
        track.selectedFormat = selectedFormat

        if self.settings['tags']['savePlaylistAsCompilation'] and track.playlist:
            track.trackNumber = track.position
            track.discNumber = "1"
            track.album = {**track.album, **track.playlist}
            ext = track.playlist['picUrl'][-4:]
            if ext[0] != ".":
                ext = ".jpg"
            track.album['picPath'] = os.path.join(TEMPDIR, f"pl{trackAPI_gw['_EXTRA_PLAYLIST']['id']}_{self.settings['embeddedArtworkSize']}{ext}")
        else:
            if track.album['date']:
                track.date = track.album['date']
            track.album['picUrl'] = "https://e-cdns-images.dzcdn.net/images/cover/{}/{}x{}-{}".format(
                track.album['pic'],
                self.settings['embeddedArtworkSize'], self.settings['embeddedArtworkSize'],
                'none-100-0-0.png' if self.settings['embeddedArtworkPNG'] else f'000000-{self.settings["jpegImageQuality"]}-0-0.jpg'
            )
            track.album['picPath'] = os.path.join(TEMPDIR, f"alb{track.album['id']}_{self.settings['embeddedArtworkSize']}{track.album['picUrl'][-4:]}")
        track.album['bitrate'] = selectedFormat

        track.dateString = formatDate(track.date, self.settings['dateFormat'])
        track.album['dateString'] = formatDate(track.album['date'], self.settings['dateFormat'])

        # Check if user wants the feat in the title
        # 0 => do not change
        # 1 => remove from title
        # 2 => add to title
        # 3 => remove from title and album title
        if self.settings['featuredToTitle'] == "1":
            track.title = track.getCleanTitle()
        elif self.settings['featuredToTitle'] == "2":
            track.title = track.getFeatTitle()
        elif self.settings['featuredToTitle'] == "3":
            track.title = track.getCleanTitle()
            track.album['title'] = track.getCleanAlbumTitle()

        # Remove (Album Version) from tracks that have that
        if self.settings['removeAlbumVersion']:
            if "Album Version" in track.title:
                track.title = re.sub(r' ?\(Album Version\)', "", track.title).strip()

        # Change Title and Artists casing if needed
        if self.settings['titleCasing'] != "nothing":
            track.title = changeCase(track.title, self.settings['titleCasing'])
        if self.settings['artistCasing'] != "nothing":
            track.mainArtist['name'] = changeCase(track.mainArtist['name'], self.settings['artistCasing'])
            for i, artist in enumerate(track.artists):
                track.artists[i] = changeCase(artist, self.settings['artistCasing'])
            for type in track.artist:
                for i, artist in enumerate(track.artist[type]):
                    track.artist[type][i] = changeCase(artist, self.settings['artistCasing'])
            track.generateMainFeatStrings()

        # Generate artist tag if needed
        if self.settings['tags']['multiArtistSeparator'] != "default":
            if self.settings['tags']['multiArtistSeparator'] == "andFeat":
                track.artistsString = track.mainArtistsString
                if track.featArtistsString and self.settings['featuredToTitle'] != "2":
                    track.artistsString += " " + track.featArtistsString
            else:
                track.artistsString = self.settings['tags']['multiArtistSeparator'].join(track.artists)
        else:
            track.artistsString = ", ".join(track.artists)

        # Generate filename and filepath from metadata
        filename = generateFilename(track, trackAPI_gw, self.settings)
        (filepath, artistPath, coverPath, extrasPath) = generateFilepath(track, trackAPI_gw, self.settings)

        if self.queueItem.cancel: raise DownloadCancelled

        # Download and cache coverart
        logger.info(f"[{track.mainArtist['name']} - {track.title}] Getting the album cover")
        track.album['picPath'] = downloadImage(track.album['picUrl'], track.album['picPath'])

        # Save local album art
        if coverPath:
            result['albumURLs'] = []
            for format in self.settings['localArtworkFormat'].split(","):
                if format in ["png","jpg"]:
                    if self.settings['tags']['savePlaylistAsCompilation'] and track.playlist:
                        if track.playlist['pic']:
                            url = "{}/{}x{}-{}".format(
                                track.album['pic'],
                                self.settings['localArtworkSize'], self.settings['localArtworkSize'],
                                'none-100-0-0.png' if format == "png" else f'000000-{self.settings["jpegImageQuality"]}-0-0.jpg'
                            )
                        else:
                            url = track.album['picUrl']
                            if format != "jpg":
                                continue
                    else:
                        url = "https://e-cdns-images.dzcdn.net/images/cover/{}/{}x{}-{}".format(
                            track.album['pic'],
                            self.settings['localArtworkSize'], self.settings['localArtworkSize'],
                            'none-100-0-0.png' if format == "png" else f'000000-{self.settings["jpegImageQuality"]}-0-0.jpg'
                        )
                    result['albumURLs'].append({'url': url, 'ext': format})
            result['albumPath'] = os.path.join(coverPath,
                                               f"{settingsRegexAlbum(self.settings['coverImageTemplate'], track.album, self.settings, trackAPI_gw['_EXTRA_PLAYLIST'] if'_EXTRA_PLAYLIST' in trackAPI_gw else None)}")

        # Save artist art
        if artistPath:
            result['artistURLs'] = []
            for format in self.settings['localArtworkFormat'].split(","):
                if format in ["png","jpg"]:
                    url = ""
                    if track.album['mainArtist']['pic'] != "":
                        url = "https://e-cdns-images.dzcdn.net/images/artist/{}/{}x{}-{}".format(
                            track.album['mainArtist']['pic'], self.settings['localArtworkSize'], self.settings['localArtworkSize'],
                            'none-100-0-0.png' if format == "png" else f'000000-{self.settings["jpegImageQuality"]}-0-0.jpg')
                    elif format == "jpg":
                        url = "https://e-cdns-images.dzcdn.net/images/artist//{}x{}-{}".format(
                            self.settings['localArtworkSize'], self.settings['localArtworkSize'], f'000000-{self.settings["jpegImageQuality"]}-0-0.jpg')
                    if url:
                        result['artistURLs'].append({'url': url, 'ext': format})
            result['artistPath'] = os.path.join(artistPath,
                f"{settingsRegexArtist(self.settings['artistImageTemplate'], track.album['mainArtist'], self.settings)}")

        # Remove subfolders from filename and add it to filepath
        if os.path.sep in filename:
            tempPath = filename[:filename.rfind(os.path.sep)]
            filepath = os.path.join(filepath, tempPath)
            filename = filename[filename.rfind(os.path.sep) + len(os.path.sep):]

        # Make sure the filepath exsists
        makedirs(filepath, exist_ok=True)
        writepath = os.path.join(filepath, filename + extensions[track.selectedFormat])

        # Save lyrics in lrc file
        if self.settings['syncedLyrics'] and track.lyrics['sync']:
            if not os.path.isfile(os.path.join(filepath, filename + '.lrc')) or self.settings['overwriteFile'] in ['y', 't']:
                with open(os.path.join(filepath, filename + '.lrc'), 'wb') as f:
                    f.write(track.lyrics['sync'].encode('utf-8'))

        trackAlreadyDownloaded = os.path.isfile(writepath)
        if not trackAlreadyDownloaded and self.settings['overwriteFile'] == 'e':
            exts = ['.mp3', '.flac', '.opus', '.m4a']
            baseFilename = os.path.join(filepath, filename)
            for ext in exts:
                trackAlreadyDownloaded = os.path.isfile(baseFilename+ext)
                if trackAlreadyDownloaded:
                    break
        if trackAlreadyDownloaded and self.settings['overwriteFile'] == 'b':
            baseFilename = os.path.join(filepath, filename)
            i = 1
            currentFilename = baseFilename+' ('+str(i)+')'+ extensions[track.selectedFormat]
            while os.path.isfile(currentFilename):
                i += 1
                currentFilename = baseFilename+' ('+str(i)+')'+ extensions[track.selectedFormat]
            trackAlreadyDownloaded = False
            writepath = currentFilename


        if extrasPath:
            if not self.extrasPath:
                self.extrasPath = extrasPath
            result['extrasPath'] = extrasPath

            # Data for m3u file
            result['filename'] = writepath[len(extrasPath):]

            # Save playlist cover
            if track.playlist:
                if not len(self.playlistURLs):
                    if track.playlist['pic']:
                        for format in self.settings['localArtworkFormat'].split(","):
                            if format in ["png","jpg"]:
                                url = "{}/{}x{}-{}".format(
                                    track.playlist['pic'],
                                    self.settings['localArtworkSize'], self.settings['localArtworkSize'],
                                    'none-100-0-0.png' if format == "png" else f'000000-{self.settings["jpegImageQuality"]}-0-0.jpg'
                                )
                                self.playlistURLs.append({'url': url, 'ext': format})
                    else:
                        self.playlistURLs.append({'url': track.playlist['picUrl'], 'ext': 'jpg'})
                if not self.playlistPath:
                    track.playlist['id'] = "pl_" + str(trackAPI_gw['_EXTRA_PLAYLIST']['id'])
                    track.playlist['genre'] = ["Compilation", ]
                    track.playlist['bitrate'] = selectedFormat
                    track.playlist['dateString'] = formatDate(track.playlist['date'], self.settings['dateFormat'])
                    self.playlistPath = f"{settingsRegexAlbum(self.settings['coverImageTemplate'], track.playlist, self.settings, trackAPI_gw['_EXTRA_PLAYLIST'])}"

            if not trackAlreadyDownloaded or self.settings['overwriteFile'] == 'y':
                logger.info(f"[{track.mainArtist['name']} - {track.title}] Downloading the track")
                track.downloadUrl = self.dz.get_track_stream_url(track.id, track.MD5, track.mediaVersion, track.selectedFormat)

                def downloadMusic(track, trackAPI_gw):
                    try:
                        with open(writepath, 'wb') as stream:
                            self.streamTrack(stream, track)
                    except DownloadCancelled:
                        remove(writepath)
                        raise DownloadCancelled
                    except (request_exception.HTTPError, DownloadEmpty):
                        remove(writepath)
                        if track.fallbackId != "0":
                            logger.warn(f"[{track.mainArtist['name']} - {track.title}] Track not available, using fallback id")
                            newTrack = self.dz.get_track_gw(track.fallbackId)
                            track.parseEssentialData(self.dz, newTrack)
                            return False
                        elif not track.searched and self.settings['fallbackSearch']:
                            logger.warn(f"[{track.mainArtist['name']} - {track.title}] Track not available, searching for alternative")
                            searchedId = self.dz.get_track_from_metadata(track.mainArtist['name'], track.title, track.album['title'])
                            if searchedId != "0":
                                newTrack = self.dz.get_track_gw(searchedId)
                                track.parseEssentialData(self.dz, newTrack)
                                track.searched = True
                                return False
                            else:
                                raise DownloadFailed("notAvailableNoAlternative")
                        else:
                            raise DownloadFailed("notAvailable")
                    except (request_exception.ConnectionError, request_exception.ChunkedEncodingError) as e:
                        remove(writepath)
                        logger.warn(f"[{track.mainArtist['name']} - {track.title}] Error while downloading the track, trying again in 5s...")
                        sleep(5)
                        return downloadMusic(track, trackAPI_gw)
                    except Exception as e:
                        remove(writepath)
                        logger.exception(str(e))
                        logger.warn(f"[{track.mainArtist['name']} - {track.title}] Error while downloading the track, you should report this to the developers")
                        raise e
                    return True

                try:
                    trackDownloaded = downloadMusic(track, trackAPI_gw)
                except DownloadFailed as e:
                    raise e
                except Exception as e:
                    raise e

                if not trackDownloaded:
                    return self.download(trackAPI_gw, track)
            else:
                logger.info(f"[{track.mainArtist['name']} - {track.title}] Skipping track as it's already downloaded")
                self.completeTrackPercentage()

            # Adding tags
            if (not trackAlreadyDownloaded or self.settings['overwriteFile'] in ['t', 'y']) and not track.localTrack:
                logger.info(f"[{track.mainArtist['name']} - {track.title}] Applying tags to the track")
                if track.selectedFormat in [3, 1, 8]:
                    tagID3(writepath, track, self.settings['tags'])
                elif track.selectedFormat == 9:
                    try:
                        tagFLAC(writepath, track, self.settings['tags'])
                    except FLACNoHeaderError:
                        remove(writepath)
                        logger.warn(f"[{track.mainArtist['name']} - {track.title}] Track not available in FLAC, falling back if necessary")
                        self.removeTrackPercentage()
                        track.filesizes['FILESIZE_FLAC'] = "0"
                        track.filesizes['FILESIZE_FLAC_TESTED'] = True
                        return self.download(trackAPI_gw, track)
                if track.searched:
                    result['searched'] = f"{track.mainArtist['name']} - {track.title}"

            logger.info(f"[{track.mainArtist['name']} - {track.title}] Track download completed\n{writepath}")
            self.queueItem.downloaded += 1
            if self.interface:
                self.interface.send("updateQueue", {'uuid': self.queueItem.uuid, 'downloaded': True, 'downloadPath': writepath})
            return result

    def getPreferredBitrate(self, track):
        if track.localTrack:
            return 0

        fallback = self.settings['fallbackBitrate']

        formats_non_360 = {
            9: "FLAC",
            3: "MP3_320",
            1: "MP3_128",
        }
        formats_360 = {
            15: "MP4_RA3",
            14: "MP4_RA2",
            13: "MP4_RA1",
        }

        if not fallback:
            error_num = -100
            formats = formats_360
            formats.update(formats_non_360)
        elif int(self.bitrate) in formats_360:
            error_num = -200
            formats = formats_360
        else:
            error_num = 8
            formats = formats_non_360

        for format_num, format in formats.items():
            if format_num <= int(self.bitrate):
                if f"FILESIZE_{format}" in track.filesizes:
                    if int(track.filesizes[f"FILESIZE_{format}"]) != 0:
                        return format_num
                    elif not track.filesizes[f"FILESIZE_{format}_TESTED"]:
                        request = get(self.dz.get_track_stream_url(track.id, track.MD5, track.mediaVersion, format_num), stream=True)
                        try:
                            request.raise_for_status()
                            return format_num
                        except request_exception.HTTPError: # if the format is not available, Deezer returns a 403 error
                            pass
                if fallback:
                    continue
                else:
                    return error_num

        return error_num # fallback is enabled and loop went through all formats

    def streamTrack(self, stream, track):
        if self.queueItem.cancel: raise DownloadCancelled

        try:
            request = get(track.downloadUrl, headers=self.dz.http_headers, stream=True, timeout=30)
        except request_exception.ConnectionError:
            sleep(2)
            return self.streamTrack(stream, track)
        request.raise_for_status()
        blowfish_key = str.encode(self.dz._get_blowfish_key(str(track.id)))
        complete = int(request.headers["Content-Length"])
        if complete == 0:
            raise DownloadEmpty
        chunkLength = 0
        percentage = 0
        i = 0
        for chunk in request.iter_content(2048):
            if self.queueItem.cancel: raise DownloadCancelled
            if i % 3 == 0 and len(chunk) == 2048:
                chunk = Blowfish.new(blowfish_key, Blowfish.MODE_CBC, b"\x00\x01\x02\x03\x04\x05\x06\x07").decrypt(chunk)
            stream.write(chunk)
            chunkLength += len(chunk)
            if isinstance(self.queueItem, QISingle):
                percentage = (chunkLength / complete) * 100
                self.downloadPercentage = percentage
            else:
                chunkProgres = (len(chunk) / complete) / self.queueItem.size * 100
                self.downloadPercentage += chunkProgres
            self.updatePercentage()
            i += 1

    def updatePercentage(self):
        if round(self.downloadPercentage) != self.lastPercentage and round(self.downloadPercentage) % 2 == 0:
            self.lastPercentage = round(self.downloadPercentage)
            self.queueItem.progress = self.lastPercentage
            if self.interface:
                self.interface.send("updateQueue", {'uuid': self.queueItem.uuid, 'progress': self.lastPercentage})

    def completeTrackPercentage(self):
        if isinstance(self.queueItem, QISingle):
            self.downloadPercentage = 100
        else:
            self.downloadPercentage += (1 / self.queueItem.size) * 100
        self.updatePercentage()

    def removeTrackPercentage(self):
        if isinstance(self.queueItem, QISingle):
            self.downloadPercentage = 0
        else:
            self.downloadPercentage -= (1 / self.queueItem.size) * 100
        self.updatePercentage()

    def downloadWrapper(self, trackAPI_gw):
        track = {
            'id': trackAPI_gw['SNG_ID'],
            'title': trackAPI_gw['SNG_TITLE'] + (trackAPI_gw['VERSION'] if 'VERSION' in trackAPI_gw and trackAPI_gw['VERSION'] and not trackAPI_gw['VERSION'] in trackAPI_gw['SNG_TITLE'] else ""),
            'artist': trackAPI_gw['ART_NAME']
        }

        try:
            result = self.download(trackAPI_gw)
        except DownloadCancelled:
            return None
        except DownloadFailed as error:
            logger.error(f"[{track['artist']} - {track['title']}] {error.message}")
            result = {'error': {
                        'message': error.message,
                        'errid': error.errid,
                        'data': track
                    }}
        except Exception as e:
            logger.exception(f"[{track['artist']} - {track['title']}] {str(e)}")
            result = {'error': {
                        'message': str(e),
                        'data': track
                    }}

        if 'error' in result:
            self.completeTrackPercentage()
            self.queueItem.failed += 1
            self.queueItem.errors.append(result['error'])
            if self.interface:
                error = result['error']
                self.interface.send("updateQueue", {
                    'uuid': self.queueItem.uuid,
                    'failed': True,
                    'data': error['data'],
                    'error': error['message'],
                    'errid': error['errid'] if 'errid' in error else None
                })
        return result

class DownloadError(Exception):
    """Base class for exceptions in this module."""
    pass

class DownloadFailed(DownloadError):
    def __init__(self, errid):
        self.errid = errid
        self.message = errorMessages[self.errid]

class DownloadCancelled(DownloadError):
    pass

class DownloadEmpty(DownloadError):
    pass



#!/usr/bin/env python3
from deemix.app.downloadjob import DownloadJob
from deemix.utils.misc import getIDFromLink, getTypeFromLink, getBitrateInt
from deemix.api.deezer import APIError
from spotipy.exceptions import SpotifyException
from deemix.app.queueitem import QISingle, QICollection, QIConvertable
import logging
import os.path as path
import json
from os import remove
from time import sleep

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('deemix')

class QueueManager:
    def __init__(self):
        self.queue = []
        self.queueList = {}
        self.queueComplete = []
        self.currentItem = ""

    def generateQueueItem(self, dz, sp, url, settings, bitrate=None, albumAPI=None, interface=None):
        forcedBitrate = getBitrateInt(bitrate)
        bitrate = forcedBitrate if forcedBitrate else settings['maxBitrate']
        type = getTypeFromLink(url)
        id = getIDFromLink(url, type)


        if type == None or id == None:
            logger.warn("URL not recognized")
            return QueueError(url, "URL not recognized", "invalidURL")

        elif type == "track":
            if id.startswith("isrc"):
                try:
                    trackAPI = dz.get_track(id)
                    if 'id' in trackAPI and 'title' in trackAPI:
                        id = trackAPI['id']
                    else:
                        return QueueError(url, "Track ISRC is not available on deezer", "ISRCnotOnDeezer")
                except APIError as e:
                    e = json.loads(str(e))
                    return QueueError(url, f"Wrong URL: {e['type']+': ' if 'type' in e else ''}{e['message'] if 'message' in e else ''}")
            try:
                trackAPI = dz.get_track_gw(id)
            except APIError as e:
                e = json.loads(str(e))
                message = "Wrong URL"
                if "DATA_ERROR" in e:
                    message += f": {e['DATA_ERROR']}"
                return QueueError(url, message)
            if albumAPI:
                trackAPI['_EXTRA_ALBUM'] = albumAPI
            if settings['createSingleFolder']:
                trackAPI['FILENAME_TEMPLATE'] = settings['albumTracknameTemplate']
            else:
                trackAPI['FILENAME_TEMPLATE'] = settings['tracknameTemplate']
            trackAPI['SINGLE_TRACK'] = True

            title = trackAPI['SNG_TITLE']
            if 'VERSION' in trackAPI and trackAPI['VERSION']:
                title += " " + trackAPI['VERSION']
            return QISingle(
                id,
                bitrate,
                title,
                trackAPI['ART_NAME'],
                f"https://e-cdns-images.dzcdn.net/images/cover/{trackAPI['ALB_PICTURE']}/75x75-000000-80-0-0.jpg",
                'track',
                settings,
                trackAPI,
            )

        elif type == "album":
            try:
                albumAPI = dz.get_album(id)
            except APIError as e:
                e = json.loads(str(e))
                return QueueError(url, f"Wrong URL: {e['type']+': ' if 'type' in e else ''}{e['message'] if 'message' in e else ''}")
            if id.startswith('upc'):
                id = albumAPI['id']
            albumAPI_gw = dz.get_album_gw(id)
            albumAPI['nb_disk'] = albumAPI_gw['NUMBER_DISK']
            albumAPI['copyright'] = albumAPI_gw['COPYRIGHT']
            if albumAPI['nb_tracks'] == 1:
                return self.generateQueueItem(dz, sp, f"https://www.deezer.com/track/{albumAPI['tracks']['data'][0]['id']}",
                                         settings, bitrate, albumAPI)
            tracksArray = dz.get_album_tracks_gw(id)
            if albumAPI['nb_tracks'] == 255:
                albumAPI['nb_tracks'] = len(tracksArray)


            if albumAPI['cover_small'] != None:
                cover = albumAPI['cover_small'][:-24] + '/75x75-000000-80-0-0.jpg'
            else:
                cover = f"https://e-cdns-images.dzcdn.net/images/cover/{albumAPI_gw['ALB_PICTURE']}/75x75-000000-80-0-0.jpg"
            totalSize = len(tracksArray)
            collection = []
            for pos, trackAPI in enumerate(tracksArray, start=1):
                trackAPI['_EXTRA_ALBUM'] = albumAPI
                trackAPI['POSITION'] = pos
                trackAPI['SIZE'] = totalSize
                trackAPI['FILENAME_TEMPLATE'] = settings['albumTracknameTemplate']
                collection.append(trackAPI)

            return QICollection(
                id,
                bitrate,
                albumAPI['title'],
                albumAPI['artist']['name'],
                cover,
                totalSize,
                'album',
                settings,
                collection,
            )


        elif type == "playlist":
            try:
                playlistAPI = dz.get_playlist(id)
            except:
                try:
                    playlistAPI = dz.get_playlist_gw(id)
                except APIError as e:
                    e = json.loads(str(e))
                    message = "Wrong URL"
                    if "DATA_ERROR" in e:
                        message += f": {e['DATA_ERROR']}"
                    return QueueError(url, message)
            if not playlistAPI['public'] and playlistAPI['creator']['id'] != str(dz.user['id']):
                logger.warn("You can't download others private playlists.")
                return QueueError(url, "You can't download others private playlists.", "notYourPrivatePlaylist")

            playlistTracksAPI = dz.get_playlist_tracks_gw(id)
            playlistAPI['various_artist'] = dz.get_artist(5080)

            totalSize = len(playlistTracksAPI)
            collection = []
            for pos, trackAPI in enumerate(playlistTracksAPI, start=1):
                if 'EXPLICIT_TRACK_CONTENT' in trackAPI and trackAPI['EXPLICIT_TRACK_CONTENT'].get('EXPLICIT_LYRICS_STATUS') in [1,4]:
                    playlistAPI['explicit'] = True
                trackAPI['_EXTRA_PLAYLIST'] = playlistAPI
                trackAPI['POSITION'] = pos
                trackAPI['SIZE'] = totalSize
                trackAPI['FILENAME_TEMPLATE'] = settings['playlistTracknameTemplate']
                collection.append(trackAPI)
            if not 'explicit' in playlistAPI:
                playlistAPI['explicit'] = False

            return QICollection(
                id,
                bitrate,
                playlistAPI['title'],
                playlistAPI['creator']['name'],
                playlistAPI['picture_small'][:-24] + '/75x75-000000-80-0-0.jpg',
                totalSize,
                'playlist',
                settings,
                collection,
            )

        elif type == "artist":
            try:
                artistAPI = dz.get_artist(id)
            except APIError as e:
                e = json.loads(str(e))
                return QueueError(url, f"Wrong URL: {e['type']+': ' if 'type' in e else ''}{e['message'] if 'message' in e else ''}")

            if interface:
                interface.send("startAddingArtist", {'name': artistAPI['name'], 'id': artistAPI['id']})

            artistAPITracks = dz.get_artist_albums(id)
            albumList = []
            for album in artistAPITracks['data']:
                albumList.append(self.generateQueueItem(dz, sp, album['link'], settings, bitrate))

            if interface:
                interface.send("finishAddingArtist", {'name': artistAPI['name'], 'id': artistAPI['id']})

            return albumList

        elif type == "artistdiscography":
            try:
                artistAPI = dz.get_artist(id)
            except APIError as e:
                e = json.loads(str(e))
                return QueueError(url, f"Wrong URL: {e['type']+': ' if 'type' in e else ''}{e['message'] if 'message' in e else ''}")

            if interface:
                interface.send("startAddingArtist", {'name': artistAPI['name'], 'id': artistAPI['id']})

            artistDiscographyAPI = dz.get_artist_discography_gw(id, 100)
            albumList = []
            for type in artistDiscographyAPI:
                if type != 'all':
                    for album in artistDiscographyAPI[type]:
                        albumList.append(self.generateQueueItem(dz, sp, album['link'], settings, bitrate))

            if interface:
                interface.send("finishAddingArtist", {'name': artistAPI['name'], 'id': artistAPI['id']})

            return albumList

        elif type == "artisttop":
            try:
                artistAPI = dz.get_artist(id)
            except APIError as e:
                e = json.loads(str(e))
                return QueueError(url, f"Wrong URL: {e['type']+': ' if 'type' in e else ''}{e['message'] if 'message' in e else ''}")

            playlistAPI = {
                'id': str(artistAPI['id'])+"_top_track",
                'title': artistAPI['name']+" - Top Tracks",
                'description': "Top Tracks for "+artistAPI['name'],
                'duration': 0,
                'public': True,
                'is_loved_track': False,
                'collaborative': False,
                'nb_tracks': 0,
                'fans': artistAPI['nb_fan'],
                'link': "https://www.deezer.com/artist/"+str(artistAPI['id'])+"/top_track",
                'share': None,
                'picture': artistAPI['picture'],
                'picture_small': artistAPI['picture_small'],
                'picture_medium': artistAPI['picture_medium'],
                'picture_big': artistAPI['picture_big'],
                'picture_xl': artistAPI['picture_xl'],
                'checksum': None,
                'tracklist': "https://api.deezer.com/artist/"+str(artistAPI['id'])+"/top",
                'creation_date': "XXXX-00-00",
                'creator': {
                    'id': "art_"+str(artistAPI['id']),
                    'name': artistAPI['name'],
                    'type': "user"
                },
                'type': "playlist"
            }

            artistTopTracksAPI_gw = dz.get_artist_toptracks_gw(id)
            playlistAPI['various_artist'] = dz.get_artist(5080)
            playlistAPI['nb_tracks'] = len(artistTopTracksAPI_gw)

            totalSize = len(artistTopTracksAPI_gw)
            collection = []
            for pos, trackAPI in enumerate(artistTopTracksAPI_gw, start=1):
                if 'EXPLICIT_TRACK_CONTENT' in trackAPI and trackAPI['EXPLICIT_TRACK_CONTENT'].get('EXPLICIT_LYRICS_STATUS') in [1,4]:
                    playlistAPI['explicit'] = True
                trackAPI['_EXTRA_PLAYLIST'] = playlistAPI
                trackAPI['POSITION'] = pos
                trackAPI['SIZE'] = totalSize
                trackAPI['FILENAME_TEMPLATE'] = settings['playlistTracknameTemplate']
                collection.append(trackAPI)
            if not 'explicit' in playlistAPI:
                playlistAPI['explicit'] = False

            return QICollection(
                id,
                bitrate,
                playlistAPI['title'],
                playlistAPI['creator']['name'],
                playlistAPI['picture_small'][:-24] + '/75x75-000000-80-0-0.jpg',
                totalSize,
                'playlist',
                settings,
                collection,
            )

        elif type == "spotifytrack":
            if not sp.spotifyEnabled:
                logger.warn("Spotify Features is not setted up correctly.")
                return QueueError(url, "Spotify Features is not setted up correctly.", "spotifyDisabled")

            try:
                track_id = sp.get_trackid_spotify(dz, id, settings['fallbackSearch'])
            except SpotifyException as e:
                return QueueError(url, "Wrong URL: "+e.msg[e.msg.find('\n')+2:])

            if track_id != "0":
                return self.generateQueueItem(dz, sp, f'https://www.deezer.com/track/{track_id}', settings, bitrate)
            else:
                logger.warn("Track not found on deezer!")
                return QueueError(url, "Track not found on deezer!", "trackNotOnDeezer")

        elif type == "spotifyalbum":
            if not sp.spotifyEnabled:
                logger.warn("Spotify Features is not setted up correctly.")
                return QueueError(url, "Spotify Features is not setted up correctly.", "spotifyDisabled")

            try:
                album_id = sp.get_albumid_spotify(dz, id)
            except SpotifyException as e:
                return QueueError(url, "Wrong URL: "+e.msg[e.msg.find('\n')+2:])

            if album_id != "0":
                return self.generateQueueItem(dz, sp, f'https://www.deezer.com/album/{album_id}', settings, bitrate)
            else:
                logger.warn("Album not found on deezer!")
                return QueueError(url, "Album not found on deezer!", "albumNotOnDeezer")

        elif type == "spotifyplaylist":
            if not sp.spotifyEnabled:
                logger.warn("Spotify Features is not setted up correctly.")
                return QueueError(url, "Spotify Features is not setted up correctly.", "spotifyDisabled")

            try:
                return sp.generate_playlist_queueitem(dz, id, bitrate, settings)
            except SpotifyException as e:
                return QueueError(url, "Wrong URL: "+e.msg[e.msg.find('\n')+2:])

        else:
            logger.warn("URL not supported yet")
            return QueueError(url, "URL not supported yet", "unsupportedURL")

    def addToQueue(self, dz, sp, url, settings, bitrate=None, interface=None):
        if not dz.logged_in:
            if interface:
                interface.send("loginNeededToDownload")
            return False

        def parseLink(link):
            link = link.strip()
            if link == "":
                return False
            logger.info("Generating queue item for: "+link)
            return self.generateQueueItem(dz, sp, link, settings, bitrate, interface=interface)

        if type(url) is list:
            queueItem = []
            for link in url:
                item = parseLink(link)
                if not item:
                    continue
                elif type(item) is list:
                    queueItem += item
                else:
                    queueItem.append(item)
            if not len(queueItem):
                return False
        else:
            queueItem = parseLink(url)
            if not queueItem:
                return False

        if type(queueItem) is list:
            ogLen = len(self.queue)
            slimmedItems = []
            for x in queueItem:
                if isinstance(x, QueueError):
                    logger.error(f"[{x.link}] {x.message}")
                    continue
                if x.uuid in list(self.queueList.keys()):
                    logger.warn(f"[{x.uuid}] Already in queue, will not be added again.")
                    continue
                self.queue.append(x.uuid)
                self.queueList[x.uuid] = x
                logger.info(f"[{x.uuid}] Added to queue.")
                slimmedItems.append(x.getSlimmedItem())
            if len(self.queue) <= ogLen:
                return False
            if interface:
                interface.send("addedToQueue", slimmedItems)
        else:
            if isinstance(queueItem, QueueError):
                logger.error(f"[{queueItem.link}] {queueItem.message}")
                if interface:
                    interface.send("queueError", queueItem.toDict())
                return False
            if queueItem.uuid in list(self.queueList.keys()):
                logger.warn(f"[{queueItem.uuid}] Already in queue, will not be added again.")
                if interface:
                    interface.send("alreadyInQueue", {'uuid': queueItem.uuid, 'title': queueItem.title})
                return False
            if interface:
                interface.send("addedToQueue", queueItem.getSlimmedItem())
            logger.info(f"[{queueItem.uuid}] Added to queue.")
            self.queue.append(queueItem.uuid)
            self.queueList[queueItem.uuid] = queueItem

        self.nextItem(dz, sp, interface)
        return True

    def nextItem(self, dz, sp, interface=None):
        if self.currentItem != "":
            return None
        else:
            if len(self.queue) > 0:
                self.currentItem = self.queue.pop(0)
            else:
                return None
            if interface:
                interface.send("startDownload", self.currentItem)
            logger.info(f"[{self.currentItem}] Started downloading.")
            DownloadJob(dz, sp, self.queueList[self.currentItem], interface).start()
            self.afterDownload(dz, sp, interface)

    def afterDownload(self, dz, sp, interface):
        if self.queueList[self.currentItem].cancel:
            del self.queueList[self.currentItem]
        else:
            self.queueComplete.append(self.currentItem)
        logger.info(f"[{self.currentItem}] Finished downloading.")
        self.currentItem = ""
        self.nextItem(dz, sp, interface)


    def getQueue(self):
        return (self.queue, self.queueComplete, self.slimQueueList(), self.currentItem)

    def saveQueue(self, configFolder):
        if len(self.queueList) > 0:
            if self.currentItem != "":
                self.queue.insert(0, self.currentItem)
            with open(path.join(configFolder, 'queue.json'), 'w') as f:
                json.dump({
                    'queue': self.queue,
                    'queueComplete': self.queueComplete,
                    'queueList': self.exportQueueList()
                }, f)

    def exportQueueList(self):
        queueList = {}
        for uuid in self.queueList:
            if uuid in self.queue:
                queueList[uuid] = self.queueList[uuid].getResettedItem()
            else:
                queueList[uuid] = self.queueList[uuid].toDict()
        return queueList

    def slimQueueList(self):
        queueList = {}
        for uuid in self.queueList:
            queueList[uuid] = self.queueList[uuid].getSlimmedItem()
        return queueList

    def loadQueue(self, configFolder, settings, interface=None):
        if path.isfile(path.join(configFolder, 'queue.json')) and not len(self.queue):
            if interface:
                interface.send('restoringQueue')
            with open(path.join(configFolder, 'queue.json'), 'r') as f:
                qd = json.load(f)
            remove(path.join(configFolder, 'queue.json'))
            self.restoreQueue(qd['queue'], qd['queueComplete'], qd['queueList'], settings)
            if interface:
                interface.send('init_downloadQueue', {
                    'queue': self.queue,
                    'queueComplete': self.queueComplete,
                    'queueList': self.slimQueueList(),
                    'restored': True
                })

    def restoreQueue(self, queue, queueComplete, queueList, settings):
        self.queue = queue
        self.queueComplete = queueComplete
        self.queueList = {}
        for uuid in queueList:
            if 'single' in queueList[uuid]:
                self.queueList[uuid] = QISingle(queueItemDict = queueList[uuid])
            if 'collection' in queueList[uuid]:
                self.queueList[uuid] = QICollection(queueItemDict = queueList[uuid])
            if '_EXTRA' in queueList[uuid]:
                self.queueList[uuid] = QIConvertable(queueItemDict = queueList[uuid])
            self.queueList[uuid].settings = settings

    def removeFromQueue(self, uuid, interface=None):
        if uuid == self.currentItem:
            if interface:
                interface.send("cancellingCurrentItem", uuid)
            self.queueList[uuid].cancel = True
        elif uuid in self.queue:
            self.queue.remove(uuid)
            del self.queueList[uuid]
            if interface:
                interface.send("removedFromQueue", uuid)
        elif uuid in self.queueComplete:
            self.queueComplete.remove(uuid)
            del self.queueList[uuid]
            if interface:
                interface.send("removedFromQueue", uuid)


    def cancelAllDownloads(self, interface=None):
        self.queue = []
        self.queueComplete = []
        if self.currentItem != "":
            if interface:
                interface.send("cancellingCurrentItem", self.currentItem)
            self.queueList[self.currentItem].cancel = True
        for uuid in list(self.queueList.keys()):
            if uuid != self.currentItem:
                del self.queueList[uuid]
        if interface:
            interface.send("removedAllDownloads", self.currentItem)


    def removeFinishedDownloads(self, interface=None):
        for uuid in self.queueComplete:
            del self.queueList[uuid]
        self.queueComplete = []
        if interface:
            interface.send("removedFinishedDownloads")

class QueueError:
    def __init__(self, link, message, errid=None):
        self.link = link
        self.message = message
        self.errid = errid

    def toDict(self):
        return {
            'link': self.link,
            'error': self.message,
            'errid': self.errid
        }



#!/usr/bin/env python3
from deemix.api.deezer import Deezer
from deemix.app.settings import Settings
from deemix.app.queuemanager import QueueManager
from deemix.app.spotifyhelper import SpotifyHelper

class deemix:
    def __init__(self, configFolder=None):
        self.set = Settings(configFolder)
        self.dz = Deezer()
        self.sp = SpotifyHelper(configFolder)
        self.qm = QueueManager()



#!/usr/bin/env python3
from mutagen.flac import FLAC, Picture
from mutagen.id3 import ID3, ID3NoHeaderError, TXXX, TIT2, TPE1, TALB, TPE2, TRCK, TPOS, TCON, TYER, TDAT, TLEN, TBPM, \
    TPUB, TSRC, USLT, APIC, IPLS, TCOM, TCOP, TCMP

# Adds tags to a MP3 file
def tagID3(stream, track, save):
    # Delete exsisting tags
    try:
        tag = ID3(stream)
        tag.delete()
    except ID3NoHeaderError:
        tag = ID3()

    if save['title']:
        tag.add(TIT2(text=track.title))

    if save['artist'] and len(track.artists):
        if save['multiArtistSeparator'] != "default":
            if save['multiArtistSeparator'] == "nothing":
                tag.add(TPE1(text=track.mainArtist['name']))
            else:
                tag.add(TPE1(text=track.artistsString))
            tag.add(TXXX(desc="ARTISTS", text=track.artists))
        else:
            tag.add(TPE1(text=track.artists))

    if save['album']:
        tag.add(TALB(text=track.album['title']))

    if save['albumArtist'] and len(track.album['artists']):
        if save['singleAlbumArtist'] and track.album['mainArtist']['save']:
            tag.add(TPE2(text=track.album['mainArtist']['name']))
        else:
            tag.add(TPE2(text=track.album['artists']))

    if save['trackNumber']:
        tag.add(TRCK(
            text=str(track.trackNumber) + ("/" + str(track.album['trackTotal']) if save['trackTotal'] else "")))
    if save['discNumber']:
        tag.add(
            TPOS(text=str(track.discNumber) + ("/" + str(track.album['discTotal']) if save['discTotal'] else "")))
    if save['genre']:
        tag.add(TCON(text=track.album['genre']))
    if save['year']:
        tag.add(TYER(text=str(track.date['year'])))
    if save['date']:
        tag.add(TDAT(text=str(track.date['month']) + str(track.date['day'])))
    if save['length']:
        tag.add(TLEN(text=str(int(track.duration)*1000)))
    if save['bpm']:
        tag.add(TBPM(text=str(track.bpm)))
    if save['label']:
        tag.add(TPUB(text=track.album['label']))
    if save['isrc']:
        tag.add(TSRC(text=track.ISRC))
    if save['barcode']:
        tag.add(TXXX(desc="BARCODE", text=track.album['barcode']))
    if save['explicit']:
        tag.add(TXXX(desc="ITUNESADVISORY", text="1" if track.explicit else "0"))
    if save['replayGain']:
        tag.add(TXXX(desc="REPLAYGAIN_TRACK_GAIN", text=track.replayGain))
    if track.lyrics['unsync'] and save['lyrics']:
        tag.add(USLT(text=track.lyrics['unsync']))

    involved_people = []
    for role in track.contributors:
        if role in ['author', 'engineer', 'mixer', 'producer', 'writer']:
            for person in track.contributors[role]:
                involved_people.append([role, person])
        elif role == 'composer' and save['composer']:
            tag.add(TCOM(text=track.contributors['composer']))
    if len(involved_people) > 0 and save['involvedPeople']:
        tag.add(IPLS(people=involved_people))

    if save['copyright']:
        tag.add(TCOP(text=track.copyright))
    if save['savePlaylistAsCompilation'] and track.playlist:
        tag.add(TCMP(text="1"))

    if save['cover'] and track.album['picPath']:
        with open(track.album['picPath'], 'rb') as f:
            tag.add(
                APIC(3, 'image/jpeg' if track.album['picPath'].endswith('jpg') else 'image/png', 3, desc='cover', data=f.read()))

    tag.save(stream, v1=2 if save['saveID3v1'] else 0, v2_version=3,
             v23_sep=None if save['useNullSeparator'] else '/')

# Adds tags to a FLAC file
def tagFLAC(stream, track, save):
    # Delete exsisting tags
    tag = FLAC(stream)
    tag.delete()
    tag.clear_pictures()

    if save['title']:
        tag["TITLE"] = track.title

    if save['artist'] and len(track.artists):
        if save['multiArtistSeparator'] != "default":
            if save['multiArtistSeparator'] == "nothing":
                tag["ARTIST"] = track.mainArtist['name']
            else:
                tag["ARTIST"] = track.artistsString
            tag["ARTISTS"] = track.artists
        else:
            tag["ARTIST"] = track.artists

    if save['album']:
        tag["ALBUM"] = track.album['title']

    if save['albumArtist'] and len(track.album['artists']):
        if save['singleAlbumArtist']:
            tag["ALBUMARTIST"] = track.album['mainArtist']['name']
        else:
            tag["ALBUMARTIST"] = track.album['artists']

    if save['trackNumber']:
        tag["TRACKNUMBER"] = str(track.trackNumber)
    if save['trackTotal']:
        tag["TRACKTOTAL"] = str(track.album['trackTotal'])
    if save['discNumber']:
        tag["DISCNUMBER"] = str(track.discNumber)
    if save['discTotal']:
        tag["DISCTOTAL"] = str(track.album['discTotal'])
    if save['genre']:
        tag["GENRE"] = track.album['genre']
    if save['date']:
        tag["DATE"] = track.dateString
    elif save['year']:
        tag["YEAR"] = str(track.date['year'])
    if save['length']:
        tag["LENGTH"] = str(track.duration)
    if save['bpm']:
        tag["BPM"] = str(track.bpm)
    if save['label']:
        tag["PUBLISHER"] = track.album['label']
    if save['isrc']:
        tag["ISRC"] = track.ISRC
    if save['barcode']:
        tag["BARCODE"] = track.album['barcode']
    if save['explicit']:
        tag["ITUNESADVISORY"] = "1" if track.explicit else "0"
    if save['replayGain']:
        tag["REPLAYGAIN_TRACK_GAIN"] = track.replayGain
    if track.lyrics['unsync'] and save['lyrics']:
        tag["LYRICS"] = track.lyrics['unsync']

    for role in track.contributors:
        if role in ['author', 'engineer', 'mixer', 'producer', 'writer', 'composer']:
            if save['involvedPeople'] and role != 'composer' or role == 'composer' and save['composer']:
                tag[role] = track.contributors[role]
        elif role == 'musicpublisher' and save['involvedPeople']:
            tag["ORGANIZATION"] = track.contributors['musicpublisher']

    if save['copyright']:
        tag["COPYRIGHT"] = track.copyright
    if save['savePlaylistAsCompilation'] and track.playlist:
        tag["COMPILATION"] = "1"

    if save['cover'] and track.album['picPath']:
        image = Picture()
        image.type = 3
        image.mime = 'image/jpeg' if track.album['picPath'].endswith('jpg') else 'image/png'
        with open(track.album['picPath'], 'rb') as f:
            image.data = f.read()
        tag.add_picture(image)

    tag.save(deleteid3=True)



#!/usr/bin/env python3
import logging

from deemix.api.deezer import APIError
from deemix.utils.misc import removeFeatures, andCommaConcat, uniqueArray

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('deemix')

class Track:
    def __init__(self, dz, settings, trackAPI_gw, trackAPI=None, albumAPI_gw=None, albumAPI=None):
        self.parseEssentialData(dz, trackAPI_gw)

        self.title = trackAPI_gw['SNG_TITLE'].strip()
        if trackAPI_gw.get('VERSION') and not trackAPI_gw['VERSION'] in trackAPI_gw['SNG_TITLE']:
            self.title += " " + trackAPI_gw['VERSION'].strip()

        self.position = trackAPI_gw.get('POSITION')

        self.localTrack = int(self.id) < 0
        if self.localTrack:
            self.parseLocalTrackData(trackAPI_gw)
        else:
            self.parseData(dz, settings, trackAPI_gw, trackAPI, albumAPI_gw, albumAPI)

        if not 'Main' in self.artist:
            self.artist['Main'] = [self.mainArtist['name']]

        # Fix incorrect day month when detectable
        if int(self.date['month']) > 12:
            monthTemp = self.date['month']
            self.date['month'] = self.date['day']
            self.date['day'] = monthTemp
        if int(self.album['date']['month']) > 12:
            monthTemp = self.album['date']['month']
            self.album['date']['month'] = self.album['date']['day']
            self.album['date']['day'] = monthTemp

        # Add playlist data if track is in a playlist
        self.playlist = None
        if "_EXTRA_PLAYLIST" in trackAPI_gw:
            self.playlist = {}
            if 'dzcdn.net' in trackAPI_gw["_EXTRA_PLAYLIST"]['picture_small']:
                self.playlist['pic'] = trackAPI_gw["_EXTRA_PLAYLIST"]['picture_small'][:-24]
                self.playlist['picUrl'] = "{}/{}x{}-{}".format(
                    self.playlist['pic'],
                    settings['embeddedArtworkSize'], settings['embeddedArtworkSize'],
                    'none-100-0-0.png' if settings['embeddedArtworkPNG'] else f'000000-{settings["jpegImageQuality"]}-0-0.jpg'
                )
            else:
                self.playlist['pic'] = None
                self.playlist['picUrl'] = trackAPI_gw["_EXTRA_PLAYLIST"]['picture_xl']
            self.playlist['title'] = trackAPI_gw["_EXTRA_PLAYLIST"]['title']
            self.playlist['mainArtist'] = {
                'id': trackAPI_gw["_EXTRA_PLAYLIST"]['various_artist']['id'],
                'name': trackAPI_gw["_EXTRA_PLAYLIST"]['various_artist']['name'],
                'pic': trackAPI_gw["_EXTRA_PLAYLIST"]['various_artist']['picture_small'][
                       trackAPI_gw["_EXTRA_PLAYLIST"]['various_artist']['picture_small'].find('artist/') + 7:-24]
            }
            if settings['albumVariousArtists']:
                self.playlist['artist'] = {"Main": [trackAPI_gw["_EXTRA_PLAYLIST"]['various_artist']['name'], ]}
                self.playlist['artists'] = [trackAPI_gw["_EXTRA_PLAYLIST"]['various_artist']['name'], ]
            else:
                self.playlist['artist'] = {"Main": []}
                self.playlist['artists'] = []
            self.playlist['trackTotal'] = trackAPI_gw["_EXTRA_PLAYLIST"]['nb_tracks']
            self.playlist['recordType'] = "Compilation"
            self.playlist['barcode'] = ""
            self.playlist['label'] = ""
            self.playlist['explicit'] = trackAPI_gw['_EXTRA_PLAYLIST']['explicit']
            self.playlist['date'] = {
                'day': trackAPI_gw["_EXTRA_PLAYLIST"]["creation_date"][8:10],
                'month': trackAPI_gw["_EXTRA_PLAYLIST"]["creation_date"][5:7],
                'year': trackAPI_gw["_EXTRA_PLAYLIST"]["creation_date"][0:4]
            }
            self.playlist['discTotal'] = "1"

        self.generateMainFeatStrings()

        # Bits useful for later
        self.searched = False
        self.selectedFormat = 0
        self.dateString = None
        self.album['picUrl'] = None
        self.album['picPath'] = None
        self.album['bitrate'] = 0
        self.album['dateString'] = None

        self.artistsString = ""

    def parseEssentialData(self, dz, trackAPI_gw):
        self.id = trackAPI_gw['SNG_ID']
        self.duration = trackAPI_gw['DURATION']
        self.MD5 = trackAPI_gw['MD5_ORIGIN']
        self.mediaVersion = trackAPI_gw['MEDIA_VERSION']
        self.fallbackId = "0"
        if 'FALLBACK' in trackAPI_gw:
            self.fallbackId = trackAPI_gw['FALLBACK']['SNG_ID']
        self.filesizes = dz.get_track_filesizes(self.id)

    def parseLocalTrackData(self, trackAPI_gw):
        self.album = {
            'id': "0",
            'title': trackAPI_gw['ALB_TITLE'],
        }
        self.album['pic'] = trackAPI_gw.get('ALB_PICTURE')
        self.mainArtist = {
            'id': "0",
            'name': trackAPI_gw['ART_NAME'],
            'pic': ""
        }
        self.artists = [trackAPI_gw['ART_NAME']]
        self.artist = {
            'Main': [trackAPI_gw['ART_NAME']]
        }
        self.date = {
            'day': "00",
            'month': "00",
            'year': "XXXX"
        }
        # All the missing data
        self.ISRC = ""
        self.album['artist'] = self.artist
        self.album['artists'] = self.artists
        self.album['barcode'] = "Unknown"
        self.album['date'] = self.date
        self.album['discTotal'] = "0"
        self.album['explicit'] = False
        self.album['genre'] = []
        self.album['label'] = "Unknown"
        self.album['mainArtist'] = self.mainArtist
        self.album['recordType'] = "Album"
        self.album['trackTotal'] = "0"
        self.bpm = 0
        self.contributors = {}
        self.copyright = ""
        self.discNumber = "0"
        self.explicit = False
        self.lyrics = {}
        self.replayGain = ""
        self.trackNumber = "0"

    def parseData(self, dz, settings, trackAPI_gw, trackAPI, albumAPI_gw, albumAPI):
        self.discNumber = trackAPI_gw.get('DISK_NUMBER')
        self.explicit = bool(int(trackAPI_gw.get('EXPLICIT_LYRICS') or "0"))
        self.copyright = trackAPI_gw.get('COPYRIGHT')
        self.replayGain = ""
        if 'GAIN' in trackAPI_gw:
            self.replayGain = "{0:.2f} dB".format((float(trackAPI_gw['GAIN']) + 18.4) * -1)
        self.ISRC = trackAPI_gw['ISRC']
        self.trackNumber = trackAPI_gw['TRACK_NUMBER']
        self.contributors = trackAPI_gw['SNG_CONTRIBUTORS']

        self.lyrics = {
            'id': trackAPI_gw.get('LYRICS_ID'),
            'unsync': None,
            'sync': None
        }
        if not "LYRICS" in trackAPI_gw and int(self.lyrics['id']) != 0:
            logger.info(f"[{trackAPI_gw['ART_NAME']} - {self.title}] Getting lyrics")
            trackAPI_gw["LYRICS"] = dz.get_lyrics_gw(self.id)
        if int(self.lyrics['id']) != 0:
            self.lyrics['unsync'] = trackAPI_gw["LYRICS"].get("LYRICS_TEXT")
            if "LYRICS_SYNC_JSON" in trackAPI_gw["LYRICS"]:
                self.lyrics['sync'] = ""
                lastTimestamp = ""
                for i in range(len(trackAPI_gw["LYRICS"]["LYRICS_SYNC_JSON"])):
                    if "lrc_timestamp" in trackAPI_gw["LYRICS"]["LYRICS_SYNC_JSON"][i]:
                        self.lyrics['sync'] += trackAPI_gw["LYRICS"]["LYRICS_SYNC_JSON"][i]["lrc_timestamp"]
                        lastTimestamp = trackAPI_gw["LYRICS"]["LYRICS_SYNC_JSON"][i]["lrc_timestamp"]
                    else:
                        self.lyrics['sync'] += lastTimestamp
                    self.lyrics['sync'] += trackAPI_gw["LYRICS"]["LYRICS_SYNC_JSON"][i]["line"] + "\r\n"

        self.mainArtist = {
            'id': trackAPI_gw['ART_ID'],
            'name': trackAPI_gw['ART_NAME'],
            'pic': trackAPI_gw.get('ART_PICTURE')
        }

        self.date = None
        if 'PHYSICAL_RELEASE_DATE' in trackAPI_gw:
            self.date = {
                'day': trackAPI_gw["PHYSICAL_RELEASE_DATE"][8:10],
                'month': trackAPI_gw["PHYSICAL_RELEASE_DATE"][5:7],
                'year': trackAPI_gw["PHYSICAL_RELEASE_DATE"][0:4]
            }

        self.album = {
            'id': trackAPI_gw['ALB_ID'],
            'title': trackAPI_gw['ALB_TITLE'],
            'pic': trackAPI_gw.get('ALB_PICTURE'),
            'barcode': "Unknown",
            'label': "Unknown",
            'explicit': False,
            'date': None,
            'genre': []
        }
        try:
            # Try the public API first (as it has more data)
            if not albumAPI:
                logger.info(f"[{self.mainArtist['name']} - {self.title}] Getting album infos")
                albumAPI = dz.get_album(self.album['id'])
            self.album['title'] = albumAPI['title']
            self.album['mainArtist'] = {
                'id': albumAPI['artist']['id'],
                'name': albumAPI['artist']['name'],
                'pic': albumAPI['artist']['picture_small'][albumAPI['artist']['picture_small'].find('artist/') + 7:-24]
            }

            self.album['artist'] = {}
            self.album['artists'] = []
            for artist in albumAPI['contributors']:
                if artist['id'] != 5080 or artist['id'] == 5080 and settings['albumVariousArtists']:
                    if artist['name'] not in self.album['artists']:
                        self.album['artists'].append(artist['name'])
                    if artist['role'] == "Main" or artist['role'] != "Main" and artist['name'] not in self.album['artist']['Main']:
                        if not artist['role'] in self.album['artist']:
                            self.album['artist'][artist['role']] = []
                        self.album['artist'][artist['role']].append(artist['name'])
            if settings['removeDuplicateArtists']:
                self.album['artists'] = uniqueArray(self.album['artists'])
                for role in self.album['artist'].keys():
                    self.album['artist'][role] = uniqueArray(self.album['artist'][role])

            self.album['trackTotal'] = albumAPI['nb_tracks']
            self.album['recordType'] = albumAPI['record_type']

            self.album['barcode'] = albumAPI.get('upc') or self.album['barcode']
            self.album['label'] = albumAPI.get('label') or self.album['label']
            self.album['explicit'] = bool(albumAPI.get('explicit_lyrics'))
            if 'release_date' in albumAPI:
                self.album['date'] = {
                    'day': albumAPI["release_date"][8:10],
                    'month': albumAPI["release_date"][5:7],
                    'year': albumAPI["release_date"][0:4]
                }
            self.album['discTotal'] = albumAPI.get('nb_disk')
            self.copyright = albumAPI.get('copyright')

            if not self.album['pic']:
                self.album['pic'] = albumAPI['cover_small'][albumAPI['cover_small'].find('cover/') + 6:-24]

            if 'genres' in albumAPI and 'data' in albumAPI['genres'] and len(albumAPI['genres']['data']) > 0:
                for genre in albumAPI['genres']['data']:
                    self.album['genre'].append(genre['name'])
        except APIError:
            if not albumAPI_gw:
                logger.info(f"[{self.mainArtist['name']} - {self.title}] Getting more album infos")
                albumAPI_gw = dz.get_album_gw(self.album['id'])
            self.album['title'] = albumAPI_gw['ALB_TITLE']
            self.album['mainArtist'] = {
                'id': albumAPI_gw['ART_ID'],
                'name': albumAPI_gw['ART_NAME'],
                'pic': None
            }
            logger.info(f"[{self.mainArtist['name']} - {self.title}] Getting artist picture fallback")
            artistAPI = dz.get_artist(self.album['mainArtist']['id'])
            self.album['artists'] = [albumAPI_gw['ART_NAME']]
            self.album['mainArtist']['pic'] = artistAPI['picture_small'][artistAPI['picture_small'].find('artist/') + 7:-24]
            self.album['trackTotal'] = albumAPI_gw['NUMBER_TRACK']
            self.album['discTotal'] = albumAPI_gw['NUMBER_DISK']
            self.album['recordType'] = "Album"
            self.album['label'] = albumAPI_gw.get('LABEL_NAME') or self.album['label']
            if 'EXPLICIT_ALBUM_CONTENT' in albumAPI_gw and 'EXPLICIT_LYRICS_STATUS' in albumAPI_gw['EXPLICIT_ALBUM_CONTENT']:
                self.album['explicit'] = albumAPI_gw['EXPLICIT_ALBUM_CONTENT']['EXPLICIT_LYRICS_STATUS'] in [1,4]
            if not self.album['pic']:
                self.album['pic'] = albumAPI_gw['ALB_PICTURE']
            if 'PHYSICAL_RELEASE_DATE' in albumAPI_gw:
                self.album['date'] = {
                    'day': albumAPI_gw["PHYSICAL_RELEASE_DATE"][8:10],
                    'month': albumAPI_gw["PHYSICAL_RELEASE_DATE"][5:7],
                    'year': albumAPI_gw["PHYSICAL_RELEASE_DATE"][0:4]
                }

        self.album['mainArtist']['save'] = self.album['mainArtist']['id'] != 5080 or self.album['mainArtist']['id'] == 5080 and settings['albumVariousArtists']

        if self.album['date'] and not self.date:
            self.date = self.album['date']

        if not trackAPI:
            logger.info(f"[{self.mainArtist['name']} - {self.title}] Getting extra track infos")
            trackAPI = dz.get_track(self.id)
        self.bpm = trackAPI['bpm']

        if not self.replayGain and 'gain' in trackAPI:
            self.replayGain = "{0:.2f} dB".format((float(trackAPI['gain']) + 18.4) * -1)
        if not self.explicit:
            self.explicit = trackAPI['explicit_lyrics']
        if not self.discNumber:
            self.discNumber = trackAPI['disk_number']

        self.artist = {}
        self.artists = []
        for artist in trackAPI['contributors']:
            if artist['id'] != 5080 or artist['id'] == 5080 and len(trackAPI['contributors']) == 1:
                if artist['name'] not in self.artists:
                    self.artists.append(artist['name'])
                if artist['role'] != "Main" and artist['name'] not in self.artist['Main'] or artist['role'] == "Main":
                    if not artist['role'] in self.artist:
                        self.artist[artist['role']] = []
                    self.artist[artist['role']].append(artist['name'])
        if settings['removeDuplicateArtists']:
            self.artists = uniqueArray(self.artists)
            for role in self.artist.keys():
                self.artist[role] = uniqueArray(self.artist[role])

        if not self.album['discTotal']:
            if not albumAPI_gw:
                logger.info(f"[{self.mainArtist['name']} - {self.title}] Getting more album infos")
                albumAPI_gw = dz.get_album_gw(self.album['id'])
            self.album['discTotal'] = albumAPI_gw['NUMBER_DISK']
        if not self.copyright:
            if not albumAPI_gw:
                logger.info(f"[{self.mainArtist['name']} - {self.title}] Getting more album infos")
                albumAPI_gw = dz.get_album_gw(self.album['id'])
            self.copyright = albumAPI_gw['COPYRIGHT']

    # Removes featuring from the title
    def getCleanTitle(self):
        return removeFeatures(self.title)

    # Removes featuring from the album name
    def getCleanAlbumTitle(self):
        return removeFeatures(self.album['title'])

    def getFeatTitle(self):
        if self.featArtistsString and not "(feat." in self.title.lower():
            return self.title + " ({})".format(self.featArtistsString)
        return self.title

    def generateMainFeatStrings(self):
        self.mainArtistsString = andCommaConcat(self.artist['Main'])
        self.featArtistsString = None
        if 'Featured' in self.artist:
            self.featArtistsString = "feat. "+andCommaConcat(self.artist['Featured'])
